#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <limits.h>
#include <selinux/selinux.h>
#include <selinux/label.h>
#include <sepol/sepol.h>

/* Get entries from $HOME/notebook.conf file */
extern void get_config_entry(char *entry, char **content);

/*
 * This routine will validate the selected file context against those
 * defined in the policy. It is set-up by the selinux_set_callback function.
 *
 * NOTE selabel_lookup is strange - Need the validate_policy routine but once
 * selabel_open has been issued and the contexts checked, selabel_lookup just
 * refers to an 'in memory' list to check if okay or not (i.e. does not call
 * validate_policy unless memory entry is invalid).
 */
int validate_policy(char **contextp)
{
    char *context = *contextp;
	int rc;

	printf("---------- Start %s callback function -----------\n", __FUNCTION__);

	if (sepol_check_context(context) < 0) {
        printf("\nsepol_check_context validated default context as "
								"INVALID:\n\t%s\n", context);
        printf("sepol_check_context - ERROR: %s\n", strerror(errno));
        rc = -EINVAL;
    }
    else {
   	    printf("sepol_check_context validated default context as VALID:"
														"\n\t%s\n", context);
    	rc = 0;
    }
	printf("---------- End %s callback function -----------\n\n", __FUNCTION__);
	return (rc);
}


/* --------------- Start Here ----------------------- */
int main (int argc, char **argv)
{
    char answer[8];
    int mode;
    struct selabel_handle *hnd;
    security_context_t selabel_context;
    char *policyfile, *path, *file_contexts_path;
    char policy_path[PATH_MAX];
    FILE *policystream;
    int ver = security_policyvers();

    /* The selabel_open options are VALIDATE contexts and use specified path */
    struct selinux_opt selabel_option[] = {
        { SELABEL_OPT_VALIDATE, (char *)1 },
        { SELABEL_OPT_PATH, file_contexts_path }
    };

#ifdef INFO
    printf("\nThe selabel_policy_file example will show the default "
				"context that should\nbe applied to a file using the selected "
				"binary policy file and associated\nfile_contexts file (and "
				"their corresponding .homedirs and .local files).\n\n");

    printf("The example requests a binary policy file name and then a "
                "file_contexts file.\nOnce selected the example will validate "
                "each file_contexts entry (because\nthe SELABEL_OPT_VALIDATE "
				"option is set), and then loop asking for a file name\nto be "
				"selected. The files default context will then be displayed."
				"\nPress 'q' to quit and call selabel_stats, or return to "
				"continue.\nPress return to continue\n");
    getchar();

    printf("This example calls the following libselinux and libsepol functions:"
                "\n  selabel_open"
                "\n     This will initialise and load the selected "
                "'file_contexts' file and\n     (if available) the .homedirs "
                "and .local files."
                "\n     Also sets SELABEL_OPT_VALIDATE to validate and "
                "\n     SELABEL_OPT_PATH to the file_contexts path."
                "\n  selinux_set_callback with SELINUX_CB_VALIDATE"
                "\n     This will set up a validation routine to allow the "
                "libsepol function\n     'sepol_check_context' to be called to "
                "obtain the file context\n     that will be applied by the "
                "policy."
                "\n  sepol_set_policydb_from_file"
                "\n     This will load the selected binary policy file."
                "\n  selabel_lookup_raw"
                "\n     This checks whether the files context is valid or not."
                "\n     The 'raw' function is used so that contexts "
				"are not translated."
                "\n  selabel_stats"
                "\n     Reports any specifications that have not been matched."
                "\n  selabel_close"
                "\n     Close the file_contexts files and free any memory."
                "\n\nPress return to continue\n");
    getchar();
#endif

    /*
     * Set the callback to our validate_policy routine. This replaces the
     * default security_check_context that would validate against the active
     * policy.
     */
    selinux_set_callback(SELINUX_CB_VALIDATE,
                                (union selinux_callback)&validate_policy);

    printf("\nSelect a binary policy file:");
    get_config_entry("[binary_policy]", &policyfile);
    if ((strcmp(policyfile, "CURRENT")) == 0)
        snprintf(policy_path, sizeof(policy_path), "%s.%d",
									selinux_binary_policy_path(), ver);
	else
		snprintf(policy_path, sizeof(policy_path), "%s", policyfile);
	free(policyfile);

	policystream = fopen(policy_path, "r");
	if (!policystream) {
        printf("ERROR - opening policy file: %s\n%s\n", policy_path,
                                                            strerror(errno));
		exit(1);
	}

    printf("Executing: sepol_set_policydb_from_file(%s);\n", policy_path);
	if (sepol_set_policydb_from_file(policystream) <  0) {
        printf("ERROR - reading policy file: %s\n%s\n", policy_path,
                                                            strerror(errno));
		exit(1);
    }
	fclose(policystream);
    printf("Policy file opened\n\n");

    /* Get the file_context location */
    printf("Now select the file_contexts location (NULL = current):");
    get_config_entry("[file_contexts_path]", &file_contexts_path);
    if ((strcmp(file_contexts_path, "NULL")) == 0)
        file_contexts_path = NULL;
    /* Set the SELABEL_OPT_PATH option to the selected file_contexts path */
	selabel_option[1].value = file_contexts_path;

    printf("Executing: selabel_open(SELABEL_CTX_FILE, selabel_option, 2)\n\n");
	if ((hnd = selabel_open(SELABEL_CTX_FILE, selabel_option, 2)) == NULL) {
	    printf("ERROR - selabel_open failed: %s\n", strerror(errno));
	    exit(1);
	}
	free(file_contexts_path);

    /* Set the file mode to '0' */
    mode = 0;

    while(1) {

        printf("Obtaining [path] entries:\n");
        get_config_entry("[path]", &path);

        printf("Executing: selabel_lookup_raw(hnd, &selabel_context, "
                                                    "%s, %d)\n", path, mode);
        /* Using the _raw function as don't want the labels translated */
        if (selabel_lookup_raw(hnd, &selabel_context, path, mode) == 0) {
            printf("The default context that will be assigned is: \n\t%s\n",
                                                            selabel_context);
	        freecon(selabel_context);
		} else {
			printf("\nselabel_lookup ERROR %s\n", strerror(errno));
			switch (errno) {
				case ENOENT:
					printf("Failed to find a default context.\n");
					break;
				case EINVAL:
					printf("Failed to validate context, or the path / mode "
															"are invalid.\n");
					break;
			}
		}
		free(path);
        printf("\nq for Quit and display selabel_stats or "
                                                    "return to continue.\n");
        fgets(answer, sizeof(answer), stdin);
        if (answer[0] == 'q') {
            printf("selabel_stats returned:\n");
            selabel_stats(hnd);
			break;
        }
	}
	selabel_close(hnd);
    exit(0);
}
